/*

Copyright (C) 2012  Observatoire de Paris

The JavaScript code in this page is free software: you can
redistribute it and/or modify it under the terms of the GNU
General Public License (GNU GPL) as published by the Free Software
Foundation, either version 3 of the License, or (at your option)
any later version.  The code is distributed WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU GPL for more details.

As additional permission under GNU GPL version 3 section 7, you
may distribute non-source (e.g., minimized or compacted) forms of
that code without the copy of the GNU GPL normally required by
section 4, provided you include this license notice and a URL
through which recipients can access the Corresponding Source.

*/

if (!Math.toRadians) {
    Math.toRadians = function(x) {
        return(x * Math.PI / 180);
    }
}
if (!Math.toDegrees) {
    Math.toDegrees = function(x) {
        return(x * 180 / Math.PI);
    }
}

var canvas;
var context;
var minJour = 1;
var maxJour = 31; // dépend du mois
var minMois = 1;
var maxMois = 12;
var minAnnee = 2000;
var maxAnnee = 2020;
var jourInitial = 1;
var moisInitial = 1;
var anneeInitiale = 2000;
var couleurEquateur = 'rgba(150, 255, 200, 0.8)';
var couleurTexteEquateur = 'rgb(100, 200, 150)';
var couleurEcliptique = 'rgba(255, 220, 150, 0.8)';
var couleurTexteEcliptique = 'rgb(220, 150, 80)';
var couleurSoleil = 'rgb(255, 255, 0)';
var couleurLambda = 'rgb(150, 100, 50)';
var couleurBeta = 'rgb(150, 50, 150)';
var couleurAlpha = 'rgb(255, 50, 50)';
var couleurDelta = 'rgb(50, 50, 255)';
var couleurEpsilon = 'rgb(50, 130, 50)';
var couleurArcs = 'lightGray';
var precisionCourbes = 100;
var imgdimx;
var imgdimy;
var sliderJour, sliderMois, sliderAnnee;
var lambda, beta, alpha, delta, epsilon;
var jour, mois, annee;
var cx, cy; // position du centre de l'image en pixels
var rayon; // rayon du grand cercle en pixels
var labelJour, labelMois, labelAnnee;
var labelInfos;
var rotObservateur; // matrice de rotation par rapport à l'observateur
var rotEcliptique; // matrice de rotation de l'écliptique par rapport à l'équateur
var rotObsEcl; // rotObservateur * rotEcliptique
var O, G, D, EA, P; // points en coordonnées sphériques
var titreLambda, titreBeta, titreAlpha, titreDelta, titreEpsilon, titreGamma, titreEA;
var idTimeout = 0;


function init() {
    canvas = document.getElementById('dessin');
    var largeur_fenetre = 'innerWidth' in window ? window.innerWidth : document.documentElement.offsetWidth;
    var hauteur_fenetre = 'innerHeight' in window ? window.innerHeight : document.documentElement.offsetHeight;
    imgdimx = Math.min(largeur_fenetre - 350, hauteur_fenetre - 50);
    if (imgdimx <= 0)
        imgdimx = 500;
    imgdimy = imgdimx;
    canvas.width = imgdimx;
    canvas.height = imgdimy;
    context = canvas.getContext('2d');
    context.translate(0.5, 0.5);
    context.font = 'bold 12px sans-serif';
    
    titreLambda = '\u03BB';
    titreBeta = '\u03B2';
    titreAlpha = '\u03B1';
    titreDelta = '\u03B4';
    titreEpsilon = '\u03B5';
    titreGamma = '\u03B3';
    
    labelInfos = document.getElementById('infos');
    
    jour = jourInitial;
    labelJour = document.getElementById('valeurJour');
    setLabelText(labelJour, 'Jour = ' + jour);
    sliderJour = new Curseur('curseurJour', minJour, maxJour, jour, maxJour - minJour + 1, false, modifCurseur);
    
    mois = moisInitial;
    labelMois = document.getElementById('valeurMois');
    setLabelText(labelMois, 'Mois = ' + mois);
    sliderMois = new Curseur('curseurMois', minMois, maxMois, mois, maxMois - minMois + 1, false, modifCurseur);
    
    annee = anneeInitiale;
    labelAnnee = document.getElementById('valeurAnnee');
    setLabelText(labelAnnee, 'Année = ' + annee);
    sliderAnnee = new Curseur('curseurAnnee', minAnnee, maxAnnee, annee, maxAnnee - minAnnee + 1, false, modifCurseur);
    
    rotObservateur = new Matrice(3, 3);
    rotEcliptique = new Matrice(3, 3);
    O = new Vecteur3D(0, 0, 0);
    G = new Vecteur3D(1, Math.PI / 2, -Math.PI / 2);
    D = new Vecteur3D(1, Math.PI / 2, Math.PI / 2);
    EA = new Vecteur3D(1, Math.PI / 2, Math.PI);
    P = new Vecteur3D(1, 0, 0);
    
    majCroquis();
}

function setLabelText(label, texte) {
    var first = label.firstChild;
    if (first == null) {
        label.appendChild(document.createTextNode(texte));
        return;
    } else
        first.nodeValue = texte;
}

function modifCurseur(e) {
    var id = e.target.id;
    if (id == 'curseurJour') {
        var valeur = sliderJour.getValeur();
        jour = valeur;
        setLabelText(labelJour, 'Jour = ' + jour);
    } else if (id == 'curseurMois') {
        var valeur = sliderMois.getValeur();
        mois = valeur;
        setLabelText(labelMois, 'Mois = ' + mois);
        maxJour = joursParMois(mois, annee);
        sliderJour.setMaximum(maxJour);
        sliderJour.setNbpas(maxJour - minJour + 1);
        sliderJour.afficher();
    } else if (id == 'curseurAnnee') {
        var valeur = sliderAnnee.getValeur();
        annee = valeur;
        setLabelText(labelAnnee, 'Année = ' + annee);
        maxJour = joursParMois(mois, annee);
        sliderJour.setMaximum(maxJour);
        sliderJour.setNbpas(maxJour - minJour + 1);
        sliderJour.afficher();
    }
    if (idTimeout != 0)
        clearTimeout(idTimeout);
    idTimeout = setTimeout(majCroquis, 0);
}

function majCroquis() {
    // calcul du temps sidéral
    // Jour Julien le 1/1/2000 à 0 h UTC : 2451544.5
    var jours = jour - 1; // jours écoulés depuis le 1/1/2000
    for (var i=2000; i<annee; i++)
        jours += joursParAn(i);
    for (var i=1; i<12; i++) {
        if (mois > i)
            jours += joursParMois(i, annee);
    }
    
    var T = (2451544.5 + jours - 2451545) / 36525.; // en siecles juliens depuis le 1/1/2000 à 12h (pour le jour donné et l'heure 0:00)
    // obliquité moyenne de l'écliptique, en degrés
    var secondes = 21.448 - T*(46.8150 + T*(0.00059 - T*(0.001813)));
    var e0 = 23.0 + (26.0 + (secondes/60.0))/60.0;
    
    var omega = Math.toRadians(125.04 - 1934.136 * T);
    // inclinaison de la Terre, en radians
    epsilon = Math.toRadians(e0 + 0.00256 * Math.cos(omega));
    
    // longitude écliptique moyenne du soleil, en degrés
    var Lm = 280.46646 + T * (36000.76983 + 0.0003032 * T);
    // anomalie moyenne du soleil, en radians
    var M1 = Math.toRadians(357.52911 + T * (35999.05029 - 0.0001537 * T));
    // équation du centre, en degrés
    var C = Math.sin(M1) * (1.914602 - T * (0.004817 + 0.000014 * T)) + Math.sin(2*M1) * (0.019993 - 0.000101 * T) + Math.sin(3*M1) * 0.000289;
    // longitude vraie, en degrés
    var lv = (Lm + C) % 360;
    // longitude apparente, en radians
    lambda = Math.toRadians(lv - 0.00569 - 0.00478 * Math.sin(omega));
    beta = 0; // approximation, si on néglige les effets de la Lune et des autres planètes
    
    var aobs = 0.2; // angle d'observation
    rotObservateur.set(0, 0, Math.cos(aobs));
    rotObservateur.set(0, 1, 0);
    rotObservateur.set(0, 2, -Math.sin(aobs));
    rotObservateur.set(1, 0, 0);
    rotObservateur.set(1, 1, 1);
    rotObservateur.set(1, 2, 0);
    rotObservateur.set(2, 0, Math.sin(aobs));
    rotObservateur.set(2, 1, 0);
    rotObservateur.set(2, 2, Math.cos(aobs));
    
    rotEcliptique.set(0, 0, 1);
    rotEcliptique.set(0, 1, 0);
    rotEcliptique.set(0, 2, 0);
    rotEcliptique.set(1, 0, 0);
    rotEcliptique.set(1, 1, Math.cos(epsilon));
    rotEcliptique.set(1, 2, Math.sin(epsilon));
    rotEcliptique.set(2, 0, 0);
    rotEcliptique.set(2, 1, -Math.sin(epsilon));
    rotEcliptique.set(2, 2, Math.cos(epsilon));
    
    rotObsEcl = rotObservateur.produitM(rotEcliptique);
    
    context.fillStyle = 'white';
    context.fillRect(0, 0, imgdimx, imgdimy);
    cx = imgdimx / 2;
    cy = imgdimy / 2;
    rayon = Math.min(cx - 20, cy - 20); // rayon de la sphère en pixels
    
    context.beginPath();
    context.fillStyle = 'rgb(230, 230, 255)';
    context.arc(cx, cy, rayon, 0, 2*Math.PI, false);
    context.fill();
    
    context.strokeStyle = '#707070';
    context.stroke();
    
    var M = new Vecteur3D(1, Math.PI / 2 - beta, lambda); // position du Soleil dans le repère de l'écliptique
    if (M.get(2) < -Math.PI)
        M.set(2, M.get(2) + 2 * Math.PI); // pour avoir la 3ème coordonnée entre -pi et pi
    var Mcart = M.spheriquesVersCartesiennes();
    var McartEq = rotEcliptique.produitV(Mcart); // rotation écliptique -> équateur
    var MsphEq = McartEq.cartesiennesVersSpheriques(); // coordonnées sphériques de M dans le repère de l'équateur
    alpha = MsphEq.get(2);
    if (alpha < 0)
        alpha += 2*Math.PI;
    delta = Math.PI/2 - MsphEq.get(1);
    var ptM = conversion(M, true);
    var ptO = new Point2D(cx, cy);
    var B = new Vecteur3D(1, Math.PI / 2, lambda); // repère écliptique
    if (B.get(2) < -Math.PI)
        B.set(2, B.get(2) + 2 * Math.PI);
    var ptB = conversion(B, true);
    var ptG = conversion(G, false);
    var K = new Vecteur3D(1, epsilon, -Math.PI / 2);
    //final Point2D ptK = conversion(K, false);
    var A = new Vecteur3D(1, Math.PI / 2, MsphEq.get(2)); // repère de l'équateur
    var pointVernal = new Vecteur3D(1, Math.PI / 2, 0); // repère écliptique ET équateur !
    
    // arc PMA
    dessinArcPMA(P, MsphEq, A, false);
    
    // arc PMB
    dessinArcPMB(P, M, B, false);
    
    dessinSoleil(M, ptM, MsphEq, false);
    
    dessinAngleBeta(B, M, false);
    
    dessinAngleDelta(A, MsphEq, false);
    
    // calcul cercle de l'écliptique (derrière)
    var pt;
    context.beginPath();
    pt = conversion(new Vecteur3D(1, Math.PI / 2, 0), true);
    context.moveTo(pt.x, pt.y);
    for (var i=1; i<=precisionCourbes; i++) {
        pt = conversion(new Vecteur3D(1, Math.PI / 2, - (i / precisionCourbes) * Math.PI), true);
        context.lineTo(pt.x, pt.y);
    }
    context.closePath();
    // affichage cercle de l'écliptique (derrière)
    context.fillStyle = couleurEcliptique;
    context.fill();
    
    dessinAngleLambda(pointVernal, B, false);
    
    // angle alpha (partie derrière l'horizon)
    if (MsphEq.get(1) > 1E-5 && MsphEq.get(1) < Math.PI - 1E-5) // étoile au pôle: alpha indéterminé
        dessinAngleAlpha(pointVernal, A, false);
    
    dessinEquateur(pointVernal);
    
    // calcul cercle de l'écliptique (devant)
    context.beginPath();
    pt = conversion(new Vecteur3D(1, Math.PI / 2, 0), true);
    context.moveTo(pt.x, pt.y);
    for (var i=1; i<=precisionCourbes/2; i++) {
        pt = conversion(new Vecteur3D(1, Math.PI / 2, (i / (precisionCourbes/2)) * Math.PI), true);
        context.lineTo(pt.x, pt.y);
    }
    context.closePath();
    // affichage cercle de l'écliptique (devant)
    context.fillStyle = couleurEcliptique;
    context.fill();
    context.fillStyle = couleurTexteEcliptique;
    if (epsilon < Math.PI/2) {
        pt = conversion(pointVernal, false);
        context.save();
        context.translate(pt.x, pt.y);
        context.rotate(-epsilon);
        context.translate(-pt.x, -pt.y);
        context.fillText('Ecliptique', pt.x - rayon * 0.75, pt.y + 10);
        context.restore();
    } else {
        pt = conversion(EA, false);
        context.save();
        context.translate(pt.x, pt.y);
        context.rotate(epsilon);
        context.translate(-pt.x, -pt.y);
        context.fillText('Ecliptique', pt.x - rayon * 0.75, pt.y - 5);
        context.restore();
    }
    
    dessinArcPMA(P, MsphEq, A, true);
    
    dessinArcPMB(P, M, B, true);
    
    dessinSoleil(M, ptM, MsphEq, true);
    
    dessinAngleLambda(pointVernal, B, true);
    
    dessinAngleBeta(B, M, true);
    
    dessinAngleEpsilon();
    
    // angle alpha (partie devant l'horizon)
    if (MsphEq.get(1) > 1E-5 && MsphEq.get(1) < Math.PI - 1E-5) // étoile au pôle: alpha indéterminé
        dessinAngleAlpha(pointVernal, A, true);
    
    dessinAngleDelta(A, MsphEq, true);
    
    // points
    context.fillStyle = 'black';
    context.strokeStyle = 'black';
    
    // affichage des points fixes
    afficherPoint(O, false, true, "O");
    //afficherPoint(G, false, true, "G");
    //afficherPoint(D, false, true, "D");
    //afficherPoint(EA, false, true, "équinoxe d'automne");
    afficherPoint(P, false, true, "P");
    
    // affichage des points en mouvement
    afficherPoint(K, false, true, "K");
    
    //afficherPoint(B, true, true, "B");
    //afficherPoint(A, false, true, "A");
    //afficherPoint(M, true, true, "M");
    afficherPoint(pointVernal, true, true, titreGamma);
    
    
    var infos = titreLambda + "=" + affichageAngleDMS(lambda, false, true);
    if (MsphEq.get(1) > 1E-5 && MsphEq.get(1) < Math.PI - 1E-5)
        infos += " \u00a0 " + titreAlpha + "=" + affichageAngleHMS(alpha);
    infos += " \u00a0 " + titreDelta + "=" + affichageAngleDMS(Math.PI/2 - MsphEq.get(1), false, false);
    setLabelText(labelInfos, infos);
    idTimeout = 0;
}

//protected void dessinEquateur(final Vecteur pointVernal) {
function dessinEquateur(pointVernal) {
    context.beginPath();
    var pt = conversion(new Vecteur3D(1, Math.PI / 2, 0), false);
    context.moveTo(pt.x, pt.y);
    for (var i=1; i<precisionCourbes/2; i++) {
        pt = conversion(new Vecteur3D(1, Math.PI / 2, (i / (precisionCourbes/2)) * (2 * Math.PI)), false);
        context.lineTo(pt.x, pt.y);
    }
    context.closePath();
    context.fillStyle = couleurEquateur;
    context.fill();
    context.fillStyle = couleurTexteEquateur;
    pt = conversion(pointVernal, false);
    context.save();
    context.translate(pt.x, pt.y);
    context.rotate(-0.1);
    context.translate(-pt.x, -pt.y);
    context.fillText('Equateur', pt.x + rayon / 4, pt.y + 20);
    context.restore();
}

//protected void dessinSoleil(final Vecteur M, final Point2D ptM, final Vecteur MsphEq, final boolean devant) {
function dessinSoleil(M, ptM, MsphEq, devant) {
    var Mdevant = (Math.cos(M.get(2)) > 0);
    if (Mdevant == devant) {
        context.beginPath();
        context.arc(ptM.x, ptM.y, 5, 0, 2*Math.PI, false);
        context.strokeStyle = 'black';
        context.stroke();
        context.fillStyle = couleurSoleil;
        context.fill();
    }
}

//protected void dessinArcPMA(final Vecteur P, final Vecteur MsphEq, final Vecteur A, final boolean devant) {
function dessinArcPMA(P, MsphEq, A, devant) {
    var Adevant = (Math.cos(A.get(2)) > 0);
    if (Adevant == devant) {
        context.strokeStyle = couleurArcs;
        var Pbis = new Vecteur3D(1, 0, MsphEq.get(2)); // P avec le même phi que SsphEq
        dessinArc(Pbis, MsphEq.get(1) < A.get(1) ? MsphEq : A, true, false, false);
    }
}

//protected void dessinArcPMB(final Vecteur P, final Vecteur M, final Vecteur B, final boolean devant) {
function dessinArcPMB(P, M, B, devant) {
    var Bdevant = (Math.cos(B.get(2)) > 0);
    if (Bdevant == devant) {
        context.strokeStyle = couleurArcs;
        var PEq = new Vecteur3D(1, 0, M.get(2)); // P dans le repère équatorial, avec le même phi que S
        dessinArc(PEq, M.get(1) < B.get(1) ? M : B, true, true, false);
    }
}

//protected void dessinAngleLambda(final Vecteur pointVernal, final Vecteur B, final boolean devant) {
function dessinAngleLambda(pointVernal, B, devant) {
    context.strokeStyle = couleurLambda;
    context.fillStyle = couleurLambda;
    var Bdevant = (Math.cos(B.get(2)) > 0);
    var Bderriere = (Math.cos(B.get(2)) <= 0);
    // début angle
    if (devant && (Bderriere || (lambda > Math.PI && Bdevant))) {
        // trait gamma-D (devant)
        dessinArc(pointVernal, D, true, true, false);
    }
    // milieu
    if (!devant && lambda > Math.PI && Bdevant) {
        // trait D-EA-G (derrière)
        dessinArc(new Vecteur3D(1, Math.PI/2, Math.PI/2), new Vecteur3D(1, Math.PI/2, 3*Math.PI/2), true, true, false);
    }
    // fin avec la flêche
    if (devant && lambda < Math.PI && Bdevant) {
        // flêche gamma-B (devant)
        dessinArc(pointVernal, B, true, true, true);
    } else if (devant && lambda > Math.PI && Bdevant) {
        // flêche G-B (devant)
        dessinArc(new Vecteur3D(1, Math.PI/2, -Math.PI/2), B, true, true, true);
    } else if (!devant && Bderriere) {
        // flêche D-B (derrière)
        dessinArc(new Vecteur3D(1, Math.PI/2, Math.PI/2), B, true, true, true);
    }
    // titre
    if (devant) {
        var moyennePhi;
        if (B.get(2) - pointVernal.get(2) < 0)
            moyennePhi = pointVernal.get(2) + (B.get(2) + 2*Math.PI - pointVernal.get(2))/2;
        else
            moyennePhi = pointVernal.get(2) + (B.get(2) - pointVernal.get(2))/2;
        var pt = conversion(new Vecteur3D(1, Math.PI/2, moyennePhi), true);
        if (epsilon < Math.PI/2)
            context.fillText(titreLambda, pt.x - 10, pt.y - 5);
        else
            context.fillText(titreLambda, pt.x, pt.y - 5);
    }
}

//protected void dessinAngleBeta(final Vecteur B, final Vecteur M, final boolean devant) {
function dessinAngleBeta(B, M, devant) {
    context.strokeStyle = couleurBeta;
    context.fillStyle = couleurBeta;
    var Bdevant = (Math.cos(B.get(2)) > 0);
    if (devant == Bdevant) {
        dessinArc(B, M, true, true, true);
        if (Math.abs(beta) > 0.1) {
            var r;
            if ((epsilon >= 0 && beta >= 0) || (epsilon < 0 && beta < 0))
                r = 1.1;
            else
                r = 0.9;
            var pt = conversion(new Vecteur3D(r, Math.PI/2 - (Math.PI/2 - M.get(1))/2, B.get(2)), true);
            context.fillText(titreBeta, pt.x - context.measureText(titreBeta).width/2, pt.y + 5);
        }
    }
}

//protected void dessinAngleEpsilon() {
function dessinAngleEpsilon() {
    context.strokeStyle = couleurEpsilon;
    context.fillStyle = couleurEpsilon;
    dessinArc(new Vecteur3D(1.01, D.get(1), D.get(2)), new Vecteur3D(1.01, Math.PI/2 - epsilon, Math.PI/2), true, false, true);
    var pt = conversion(new Vecteur3D(1.05, Math.PI/2 - epsilon, Math.PI/2), false);
    context.fillText(titreEpsilon, pt.x - context.measureText(titreEpsilon).width/2, pt.y + 5);
}

//protected void dessinAngleAlpha(final Vecteur pointVernal, final Vecteur A, final boolean devant) {
function dessinAngleAlpha(pointVernal, A, devant) {
    var Adevant = (Math.cos(A.get(2)) > 0);
    var Aderriere = (Math.cos(A.get(2)) <= 0);
    context.strokeStyle = couleurAlpha;
    context.fillStyle = couleurAlpha;
    // début angle
    if (devant && (Aderriere || (alpha > Math.PI && Adevant))) {
        // trait gamma-D (devant)
        dessinArc(pointVernal, D, true, false, false);
    }
    // milieu
    if (!devant && alpha > Math.PI && Adevant) {
        // trait D-EA-G (derrière)
        dessinArc(new Vecteur3D(1, Math.PI/2, Math.PI/2), new Vecteur3D(1, Math.PI/2, 3*Math.PI/2), true, false, false);
    }
    // fin avec la flêche
    if (devant && alpha < Math.PI && Adevant) {
        // flêche gamma-A (devant)
        dessinArc(pointVernal, A, true, false, true);
    } else if (devant && alpha > Math.PI && Adevant) {
        // flêche G-A (devant)
        dessinArc(new Vecteur3D(1, Math.PI/2, -Math.PI/2), A, true, false, true);
    } else if (!devant && Aderriere) {
        // flêche D-A (derrière)
        dessinArc(new Vecteur3D(1, Math.PI/2, Math.PI/2), A, true, false, true);
    }
    
    if (devant) {
        var pt = conversion(new Vecteur3D(1, Math.PI/2, 0.75*alpha), false);
        context.fillText(titreAlpha, pt.x - context.measureText(titreAlpha).width/2, pt.y - 5);
    }
}

//protected void dessinAngleDelta(final Vecteur A, final Vecteur MsphEq, final boolean devant) {
function dessinAngleDelta(A, MsphEq, devant) {
    var Adevant = (Math.cos(A.get(2)) > 0);
    if (Adevant == devant) {
        context.strokeStyle = couleurDelta;
        context.fillStyle = couleurDelta;
        dessinArc(A, MsphEq, true, false, true);
        if (Math.abs(MsphEq.get(1) - Math.PI/2) > 0.1) {
            var pt = conversion(new Vecteur3D(1, Math.PI/2 - (Math.PI/2 - MsphEq.get(1))/2, A.get(2)), false);
            var x;
            if (beta >= 0)
                x = pt.x - context.measureText(titreDelta).width - 5;
            else
                x = pt.x + 5;
            context.fillText(titreDelta, x, pt.y + 5);
        }
    }
}


// dessin d'un arc entre 2 points dans l'un des 2 systèmes de coordonnées sphériques, en suivant l'une des coordonnées
//protected void dessinArc(final Vecteur v1, final Vecteur v2, final boolean sensPhiPositif, final boolean coordEcl, final boolean fleche)
function dessinArc(v1, v2, sensPhiPositif, coordEcl, fleche) {
    var nb = Math.floor(precisionCourbes / 4);
    var dr = (v2.get(0) - v1.get(0)) / nb;
    var dtheta = (v2.get(1) - v1.get(1)) / nb;
    var phi1 = v1.get(2);
    var phi2 = v2.get(2);
    if (v1.get(1) == 0)
        phi1 = phi2;
    else if (v2.get(1) == 0)
        phi2 = phi1;
    if (phi1 < 0)
        phi1 += 2 * Math.PI;
    if (phi2 < 0)
        phi2 += 2 * Math.PI;
    if (sensPhiPositif && phi1 > phi2)
        phi2 += 2 * Math.PI;
    else if (!sensPhiPositif && phi1 < phi2)
        phi1 += 2 * Math.PI;
    var dphi = (phi2 - phi1) / nb; // ne marche pas toujours (par ex. avec theta = 0)
    context.beginPath();
    var pt = conversion(v1, coordEcl);
    var vpt = null;
    context.moveTo(pt.x, pt.y);
    for (var i=1; i<=nb; i++) {
        if (i == nb)
            vpt = pt;
        pt = conversion(new Vecteur3D(v1.get(0) + i * dr, v1.get(1) + i * dtheta, phi1 + i * dphi), coordEcl);
        context.lineTo(pt.x, pt.y);
    }
    context.stroke();
    if (fleche && Math.abs(v2.get(0) - v1.get(0)) + Math.abs(v2.get(1) - v1.get(1)) + Math.abs(phi2 - phi1) > 0)
        dessinFleche(vpt, pt);
}

// dessin des 2 traits de la fin de la flêche à partir des 2 derniers points d'un arc
//protected void dessinFleche(final Point2D pt1, final Point2D pt2)
function dessinFleche(pt1, pt2) {
    var l = 6; // longueur des traits en pixels
    var da = Math.PI / 8; // angle des traits avec la fin de l'arc
    var dx = pt2.x - pt1.x;
    var dy = pt2.y - pt1.y;
    var angle;
    if (dx == 0) {
        if (dy > 0)
            angle = Math.PI / 2;
        else
            angle = -Math.PI / 2;
    } else {
        angle = Math.atan(dy / dx);
        if (dx < 0)
            angle = angle + Math.PI;
    }
    var ptFleche = new Point2D(pt2.x - l * Math.cos(angle + da), pt2.y - l * Math.sin(angle + da));
    context.beginPath();
    context.moveTo(pt2.x, pt2.y);
    context.lineTo(ptFleche.x, ptFleche.y);
    context.stroke();
    ptFleche = new Point2D(pt2.x - l * Math.cos(angle - da), pt2.y - l * Math.sin(angle - da));
    context.beginPath();
    context.moveTo(pt2.x, pt2.y);
    context.lineTo(ptFleche.x, ptFleche.y);
    context.stroke();
}

// passage des coordonnées sphériques aux coordonnées écran
//protected Point2D conversion(final Vecteur vsph, final boolean rotEq)
function conversion(vsph, rotEq) {
    // passage aux coordonnées cartésiennes
    var v = vsph.spheriquesVersCartesiennes();
    
    // rotation optionelle de l'équateur et rotation permettant de voir l'horizon de haut
    if (rotEq)
        v = rotObsEcl.produitV(v);
    else
        v = rotObservateur.produitV(v);
    
    return(projectionEcran(v));
}

// ajout d'un point en coordonnées sphériques (r, theta, phi)
//protected void afficherPoint(final Vecteur vsph, final boolean rotEq, final boolean croix, final String titre)
function afficherPoint(vsph, rotEq, croix, titre) {
    var p = conversion(vsph, rotEq);
    
    // affichage d'une croix ou d'un point
    if (croix) {
        context.beginPath();
        context.moveTo(p.x - 2, p.y);
        context.lineTo(p.x + 2, p.y);
        context.stroke();
        context.beginPath();
        context.moveTo(p.x, p.y - 2);
        context.lineTo(p.x, p.y + 2);
        context.stroke();
    } else
        context.beginPath();
        context.moveTo(p.x, p.y);
        context.lineTo(p.x, p.y);
        context.stroke();
    
    // affichage du titre
    if (titre != null) {
        var l = context.measureText(titre).width;
        context.fillText(titre, p.x - l/2, p.y - 9);
    }
}

// projection selon l'axe x à partir des coordonnées cartésiennes, puis conversion vers l'écran
//protected Point2D projectionEcran(final Vecteur v)
function projectionEcran(v) {
    return(new Point2D(cx + v.get(1) * rayon, cy - v.get(2) * rayon));
}

// affichage en heures minutes d'un angle en radians > 0
//protected static String affichageAngleHMS(final double a)
function affichageAngleHMS(a) {
    var h = a*12/Math.PI;
    var ih = Math.floor(h);
    var m = (h - ih)*60;
    var im = Math.round(m);
    if (im >= 60) {
        im -= 60;
        ih += 1;
    }
    return(ih + 'h ' + im + 'm');
}

// affichage en degrés minutes d'un angle en radians
//protected static String affichageAngleDMS(final double a, final boolean estOuest, final boolean angle360)
function affichageAngleDMS(a, estOuest, angle360) {
    var d = Math.toDegrees(Math.abs(a));
    var id = Math.floor(d);
    var m = (d - id)*60;
    var im = Math.round(m);
    if (im >= 60) {
        im -= 60;
        id += 1;
    }
    var signe;
    if (angle360 || a == 0)
        signe = '';
    else if (a > 0)
        signe = '+';
    else
        signe = '-';
    if (estOuest)
        return(id + '° ' + im + "'" + (a >= 0 ? ' E' : ' O'));
    else
        return(signe + id + '° ' + im + "'");
}

//protected static int joursParMois(final int mois, final int an)
function joursParMois(mois, an) {
    switch (mois) {
        case 1: return(31);
        case 2: {
            if (an % 4 == 0 && (an % 100 != 0 || an % 400 == 0))
                return(29);
            else
                return(28);
        }
        case 3: return(31);
        case 4: return(30);
        case 5: return(31);
        case 6: return(30);
        case 7: return(31);
        case 8: return(31);
        case 9: return(30);
        case 10: return(31);
        case 11: return(30);
        case 12: return(31);
    }
    return(0);
}

//protected static int joursParAn(final int an)
function joursParAn(an) {
    if (an % 4 == 0 && (an % 100 != 0 || an % 400 == 0))
        return(366);
    else
        return(365);
}


function Matrice(lignes, colonnes) {
    // m = new double[colonnes][lignes]
    this.m = new Array(colonnes);
    for (var i=0; i<colonnes; i++)
        this.m[i] = new Array(lignes);
    return(this);
}
//public Vecteur produit(final Vecteur v)
Matrice.prototype.produitV = function(v) {
    var nbc = this.m.length;
    var nbl = this.m[0].length;
    if (nbc != v.getDimensions()) {
        alert("produit impossible d'une matrice avec un vecteur (" + nbc + " != " + v.getDimensions() + ")");
        return(null);
    }
    var res = new Vecteur(nbl);
    var d;
    for (var j=0; j<nbl; j++) {
        d = 0;
        for (var i=0; i<nbc; i++)
            d += v.get(i) * this.m[i][j];
        res.set(j, d);
    }
    return(res);
}
//public Matrice produit(final Matrice mat)
Matrice.prototype.produitM = function(mat) {
    var nbc = this.m.length;
    var nbl = this.m[0].length;
    if (nbc != mat.getLignes()) {
        alert("produit impossible d'une matrice avec une matrice (" + nbc + " != " + mat.getLignes() + ")");
        return(null);
    }
    var res = new Matrice(nbl, nbc);
    var d;
    for (var jr=0; jr<nbc; jr++) {
        for (var ir=0; ir<nbl; ir++) {
            d = 0;
            for (var i=0; i<nbc; i++)
                d += this.m[i][jr] * mat.get(ir, i);
            res.set(ir, jr, d);
        }
    }
    return(res);
}
//public int getColonnes()
Matrice.prototype.getColonnes = function() {
    return(this.m.length);
}
//public int getLignes()
Matrice.prototype.getLignes = function() {
    return(this.m[0].length);
}
//public double get(final int colonne, final int ligne)
Matrice.prototype.get = function(colonne, ligne) {
    return(this.m[colonne][ligne]);
}
//public void set(final int colonne, final int ligne, final double valeur)
Matrice.prototype.set = function(colonne, ligne, valeur) {
    this.m[colonne][ligne] = valeur;
}


//public Vecteur(final int dimensions)
function Vecteur(dimensions) {
    // v = new double[dimensions];
    if (typeof(dimensions) == 'undefined')
        this.v = null;
    else
        this.v = new Array(dimensions);
    return(this);
}
//public Vecteur(final double v1, final double v2, final double v3)
Vecteur3D.prototype = new Vecteur();
Vecteur3D.prototype.constructor = Vecteur3D;
function Vecteur3D(v1, v2, v3) {
    Vecteur.call(this, 3);
    this.v = [v1, v2, v3];
    return(this);
}
//public int getDimensions()
Vecteur.prototype.getDimensions = function() {
    return(this.v.length);
}
//public double get(final int pos)
Vecteur.prototype.get = function(pos) {
    return(this.v[pos]);
}
//public void set(final int pos, final double valeur)
Vecteur.prototype.set = function(pos, valeur) {
    this.v[pos] = valeur;
}
// coordonnées sphériques (r, theta, phi) vers cartésiennes (x, y, z)
//public Vecteur spheriquesVersCartesiennes()
Vecteur.prototype.spheriquesVersCartesiennes = function() {
    if (this.v.length != 3) {
        alert("Conversion impossible de coordonnées (" + this.v.length + " != 3)");
        return(null);
    }
    var res = new Vecteur(3);
    res.set(0, this.v[0] * Math.sin(this.v[1]) * Math.cos(this.v[2]));
    res.set(1, this.v[0] * Math.sin(this.v[1]) * Math.sin(this.v[2]));
    res.set(2, this.v[0] * Math.cos(this.v[1]));
    return(res);
}
// coordonnées cartésiennes (x, y, z) vers sphériques (r, theta, phi)
//public Vecteur cartesiennesVersSpheriques()
Vecteur.prototype.cartesiennesVersSpheriques = function() {
    if (this.v.length != 3) {
        System.err.println("Conversion impossible de coordonnées (" + this.v.length + " != 3)");
        return(null);
    }
    var res = new Vecteur(3);
    var r = Math.sqrt(this.v[0]*this.v[0] + this.v[1]*this.v[1] + this.v[2]*this.v[2]);
    res.set(0, r);
    res.set(1, Math.acos(this.v[2]/r));
    if (this.v[0] > 0) // x > 0
        res.set(2, Math.atan(this.v[1]/this.v[0]));
    else if (this.v[0] < 0) // x < 0
        res.set(2, Math.PI + Math.atan(this.v[1]/this.v[0]));
    else if (this.v[1] > 0) // x = 0, y > 0
        res.set(2, Math.PI / 2);
    else if (this.v[1] < 0) // x = 0, y < 0
        res.set(2, 3 * Math.PI / 2);
    else // x = 0, y = 0 : peu importe !
        res.set(2, 0);
    return(res);
}

function Point2D(x, y) {
    this.x = x;
    this.y = y;
}

